/*
 *  Copyright 2006 Columbia University.
 *
 *  This file is part of MEAPsoft.
 *
 *  MEAPsoft is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 2 as
 *  published by the Free Software Foundation.
 *
 *  MEAPsoft is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with MEAPsoft; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 *  02110-1301 USA
 *
 *  See the file "COPYING" for the text of the license.
 */

package com.meapsoft;

import gnu.getopt.Getopt;

import java.io.IOException;
import java.util.Iterator;
import java.util.Vector;

import javax.sound.sampled.AudioInputStream;

import com.meapsoft.featextractors.AvgMelSpec;
import com.meapsoft.featextractors.FeatureExtractor;

/**
 * Program that extracts features from the chunks listed in the input
 * files. 
 *
 * @author Ron Weiss (ronw@ee.columbia.edu)
 */
public class FeatExtractor extends MEAPUtil
{
    // Files to use
    FeatFile[] featFiles;
    FeatFile outFile = null;
    
    // all of our feature extractors...
    Vector featExts;
    String feat_names = "";

    /**
     * FeatExtractor constructor.  If no extractors is empty, defaults
     * to AvgMelSpec.
     */
    public FeatExtractor(String infile, String outfile, Vector extractors)
    {
        this(new FeatFile(infile), new FeatFile(outfile), extractors);
    }

    /**
     * FeatExtractor constructor.  If no extractors is empty, defaults
     * to AvgMelSpec.
     */
    public FeatExtractor(FeatFile infile, FeatFile outfile, Vector extractors)
    {
        this((FeatFile[])null, outfile, extractors);       
        
        featFiles = new FeatFile[1];
        featFiles[0] = infile;
    }

    /**
     * FeatExtractor constructor.  If no extractors is empty, defaults
     * to AvgMelSpec.
     */
    public FeatExtractor(FeatFile[] infiles, FeatFile outfile, Vector extractors)
    {
        featFiles = infiles;
        outFile = outfile;
        featExts = extractors;

        if(extractors.size() == 0)
            extractors.add(new AvgMelSpec());
    }

    public void printUsageAndExit() 
    {
        System.out.println("Usage: FeatExtractor [-options] file1.feat file2.feat ... \n\n" + 
             "  where options include:\n" + 
             "    -o output_file  append features into output file (defaults to input file)");
        printCommandLineOptions('f');
        System.out.println();
        System.exit(0);
    }

    /**
     * FeatExtractor constructor.  Parses command line arguments
     */
    public FeatExtractor(String[] args) 
    {
        if(args.length == 0)
            printUsageAndExit();
        
        // Parse arguments
        String argString = "f:o:";
        featExts = parseFeatureExtractor(args, argString);

        Getopt opt = new Getopt("FeatExtracter", args, argString);
        opt.setOpterr(false);
        
        int c = -1;
        while ((c = opt.getopt()) != -1) 
        {
            switch(c) 
            {
            case 'o':
                outFile = new FeatFile(opt.getOptarg());
                break;
            case 'f': // already handled above
                break;
            case '?':
                printUsageAndExit();
                break;
            default:
                System.out.print("getopt() returned " + c + "\n");
            }
        }
        
        // parse arguments
        int ind = opt.getOptind();
        if(ind > args.length)
            printUsageAndExit();

        featFiles = new FeatFile[args.length - ind];
        for(int i=ind; i<args.length; i++)
            featFiles[i-ind] = new FeatFile(args[i]);
        
        // What are the feature names?
        for(int i = 0; i < featExts.size(); i++)
            feat_names += featExts.get(i).getClass().getName() + " ";
        //chop off last space
        feat_names = feat_names.substring(0, feat_names.length()-1);
    }

    public void setup() throws IOException, ParserException
    {
        for(int i = 0; i < featFiles.length; i++) 
            if(!featFiles[i].haveReadFile)
                featFiles[i].readFile();
    }
 

    /**
     * Where the magic happens.  Extract features from featFiles.     
     */
    public FeatFile[] processFeatFiles() throws IOException
    {
        for(int i = 0; i < featFiles.length; i++) 
            processFeatFile(featFiles[i]);

        return featFiles;
    }

    /**
     * Where the magic happens.  Extract features from file.
     */
    public FeatFile processFeatFile(FeatFile f) throws IOException
    {
        FeatFile file = (FeatFile)f.clone();

        STFT stft = null;

        file.featureDescription = "";
        boolean wroteFeatDesc = false;
        
        String lastAudioFile="";
        Iterator c = file.chunks.iterator();
        while(c.hasNext())
        {
            FeatChunk ch = (FeatChunk)c.next();

            // lets get some new features
            // compute the STFT of the whole file if this is the first 
            // time we have seen it
            if(!ch.srcFile.equals(lastAudioFile))
            {
                AudioInputStream ais = openInputStream(ch.srcFile);

                // store STFT of the whole file
                int histSize = (int)(ais.getFrameLength()/frameSize); 

                // weird bug when reading in some MP3 files -
                // ais.getFrameLength() returns -1 for some reason.
                // Hardcode a history of 10 minutes for this (I hope
                // you have a lot of RAM).  Hopefully this is big
                // enough to hold the entire file - problems will
                // occur when reading in files that are longer than 10
                // minutes and are affected by this bug
                if(histSize <= 0)
                {
                    histSize = 10*60*samplingRate/frameSize;
                    System.out.println("Warning - problem reading audio file... guessing file length.  If this doesn't "
                                       + "work like you expected it to, try again with a 44.1kHz mono WAV file.");
                }

                stft = new STFT(ais, frameSize, histSize);
                stft.start();
                ais.close();
            }
            lastAudioFile = ch.srcFile;

            // compute features from the STFT
            Iterator i = featExts.listIterator();
            while(i.hasNext())
            {
                FeatureExtractor fe = (FeatureExtractor)i.next();
                double[] feats = fe.features(stft, stft.seconds2fr(ch.startTime), 
                                             (int)stft.seconds2fr(ch.length));
                ch.addFeature(feats);

                // what features are we adding?  
                if(!wroteFeatDesc)
                    file.featureDescription += fe.getClass().getName() + "(" + feats.length + ") ";
            }

            wroteFeatDesc = true;
        }

        if(outFile != null)
        {
            outFile.chunks.addAll(file.chunks);
            outFile.featureDescription = file.featureDescription;
            
            // outFile now contains some chunks.
            outFile.haveReadFile = true;
        }

        return file;
    }
    

    /**
     * Set everything up, process input, and write output.
     */
    public void go() 
    {
        try
        {
            setup();
        }
        catch(Exception e) 
        {
            e.printStackTrace();
            System.exit(1);
        }

        FeatFile fn = outFile;

        // process supplied files
        for(int i = 0; i < featFiles.length; i++) 
        {
            if(outFile == null)
                fn = featFiles[i];
            System.out.println("Extracting features (" + feat_names + 
                               ") from " + featFiles[i].filename + " to " 
                               + fn.filename + "."); 
            
            long startTime = System.currentTimeMillis();
            try
            {
                FeatFile f = processFeatFile(featFiles[i]);
                f.writeFile(fn.filename);
            }
            catch(Exception e) 
            {
                e.printStackTrace();
                System.exit(1);
            }
            System.out.println("Done.  Took " + 
                               ((System.currentTimeMillis() - startTime)/1000.0)
                               + "s");
        }
    }
    
    
    public static void main(String[] args) 
    {
        FeatExtractor o2or = new FeatExtractor(args);
        o2or.go();
        System.exit(0);
    }
}

