/*
 *  Copyright 2006 Columbia University.
 *
 *  This file is part of MEAPsoft.
 *
 *  MEAPsoft is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 2 as
 *  published by the Free Software Foundation.
 *
 *  MEAPsoft is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with MEAPsoft; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 *  02110-1301 USA
 *
 *  See the file "COPYING" for the text of the license.
 */

package com.meapsoft;

import java.io.BufferedReader;
import java.io.FileReader;
import java.io.IOException;
import java.io.Writer;
import java.util.Iterator;
import java.util.Vector;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * Representation of a MEAPsoft segment/feature file.
 *
 * @author Ron Weiss (ronw@ee.columbia.edu)
 */

public class FeatFile extends MEAPFile implements Cloneable
{
    String featureDescription = "";
    public Vector chunks;  // Vector of FeatChunks
    
    // regular expressions for parsing FEATFiles
    protected static final Pattern commentPattern = Pattern.compile("#\\.*");
    protected static final Pattern linePattern = Pattern.compile("\\s*([^#\\s]+)\\s*");
    protected static final Pattern featDescPattern = Pattern.compile("^#\\s*Features:\\s*");

    public FeatFile(String fn)
    {
        filename = fn;
        chunks = new Vector(100, 0);
    }

    // Java bitches if this is not present for some reason.  Don't use
    // it - its bad news.
    protected FeatFile() { }


    /**
     * Get a matrix (2D double array) of all of the features in all of
     * the chunks contained in this file.  Each column corresponds to
     * a single feature vector.  Index it as you would in Matlab
     * ([row][column] = [featdim][chunk]).
     */
    public double[][] getFeatures()
    {
        return getFeatures(null);
    }

    /**
     * Get a matrix (2D double array) of all of the features in all of
     * the chunks contained in this file.  Each column corresponds to
     * a single feature vector.  Index it as you would in Matlab
     * ([row][column] = [featdim][chunk]).
     */
    public double[][] getFeatures(int[] featdim)
    {
        // how many feature dimensions are we using?
        int maxdim = 0;
        if(featdim != null)
            maxdim = featdim.length;
        else
            maxdim = ((FeatChunk)chunks.get(0)).numFeatures();

        double[][] mat = new double[chunks.size()][maxdim];

        for(int x = 0; x < chunks.size(); x++)
        {
            FeatChunk c = (FeatChunk)chunks.get(x);

            double[] currFeat = c.getFeatures(featdim);
            for(int y = 0; y < currFeat.length; y++)
                mat[x][y] = currFeat[y];
        }
        
        return mat;
    }

    /**
     * Read in an feature file
     */
    public void readFile() throws IOException, ParserException
    {
        BufferedReader in  = new BufferedReader(new FileReader(filename));

        String audioFile;
        double chunkStartTime;
        double chunkLength;

        // each line (excluding comments) should look like: 
        // audioFile chunkStartTime chunkLength feature1 feature2 ...

        // Parse each line of the input file            
        boolean haveWrittenHeader = false;
        long lineno = 0;
        String line;
        while((line = in.readLine()) != null) 
        { 
            lineno++;
            
            // extract any comments from the current line
            String comment = "";
            Matcher c = commentPattern.matcher(line+"\n");
            if(c.find())
            {
                // comments go all the way to the end of the line
                comment = c.group() + line.substring(c.end()) + "\n";
                line = line.substring(0, c.start());

                // extract featureDescription from comment
                Matcher fd = featDescPattern.matcher(comment);
                if(fd.find())
                    featureDescription = comment.substring(c.end()).trim();
            }

            Matcher p = linePattern.matcher(line);
            // is there anything else?
            if(!p.find())
                continue;
            audioFile = p.group(1);
            if(!p.find())
                throw new ParserException(filename, lineno, 
                                          "Could not find chunk start time.");
            try { chunkStartTime = Double.parseDouble(p.group(1)); }
            catch(NumberFormatException nfe) { 
                throw new ParserException(filename, lineno, 
                                          "Could not parse chunk start time \"" 
                                          + p.group(1)  + "\".");  }
            if(!p.find())
                throw new ParserException(filename, lineno, 
                                          "Could not find chunk length.");
            try { chunkLength = Double.parseDouble(p.group(1)); }
            catch(NumberFormatException nfe) { 
                throw new ParserException(filename, lineno, 
                                          "Could not parse chunk length \"" 
                                          + p.group(1)  + "\".");  }
            
            FeatChunk ch = new FeatChunk(audioFile, chunkStartTime, chunkLength);
            ch.comment = comment;

            // save the remaining features on the line
            while(p.find())
            {
                // what kind of feature is this?  If its not a double then its a string;
                try 
                {
                    ch.addFeature(Double.parseDouble(p.group(1)));
                }
                catch (NumberFormatException e)
                {
                    ch.addFeature(p.group(1));
                }
            }

            chunks.add(ch);
        }

        in.close();
        haveReadFile = true;
    }
    
    /**
     *  Remove any chunks in this file
     */
    public void clearChunks()
    {
        chunks.clear();
    }

    /**
     *  Remove the features of all chunks in this file
     */
    public void clearFeatures()
    {
        Iterator i = chunks.iterator();
        while(i.hasNext())
            ((FeatChunk)i.next()).clearFeatures();
    }
    
    /**
     * Write the contents of this FeatFile
     */
    protected void write(Writer w) throws IOException
    {
        // write the header
        w.write("# filename onset_time chunk_length [features]\n# Features: ");
        w.write(featureDescription + "\n");

        Iterator i = chunks.iterator();
        while(i.hasNext())
            w.write(i.next().toString());
    }

    /**
     * Clone this FeatFile
     */
    public Object clone()
    {
        FeatFile o = new FeatFile(this.filename);
       
        // superclass (MEAPFile) fields
        o.haveReadFile = this.haveReadFile;
        o.haveWrittenFile = this.haveWrittenFile;

        // local fields
        o.featureDescription = this.featureDescription;

        o.chunks = new Vector();
        Iterator i = this.chunks.iterator();
        while(i.hasNext())
            o.chunks.add(((FeatChunk)i.next()).clone());

        return o;
    }

    /**
     *  Check if another FeatFile is compatible with this one.  Two
     *  feature files are said to be incompatible if they contain
     *  different features.
     */
    public boolean isCompatibleWith(FeatFile f)
    {
        //todo{Should really check if feature names match excluding
        //whitespace and other meaningless gunk.}
        return this.featureDescription.equalsIgnoreCase(f.featureDescription);
    }
}
