/*
 *  Copyright 2006 Columbia University.
 *
 *  This file is part of MEAPsoft.
 *
 *  MEAPsoft is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 2 as
 *  published by the Free Software Foundation.
 *
 *  MEAPsoft is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with MEAPsoft; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 *  02110-1301 USA
 *
 *  See the file "COPYING" for the text of the license.
 */

package com.meapsoft.featextractors;

import java.util.Arrays;

import com.meapsoft.MEAPUtil;
import com.meapsoft.STFT;

/**
 * Averages all spectral frames together into a single feature vector
 * and then converts the vector to the mel frequency scale.
 *
 * @author Mike Mandel (mim@ee.columbia.edu)
 */
public class AvgMelSpec extends FeatureExtractor {
	
  double[] centersMel;
  int[] centersInd;
  double[] freq2mel;
  double[] linSpec;
  int N, outDim;

  public double lin2mel(double fq) {
    return 1127.0 * Math.log(1.0 + fq / 700.0);
  }

  public double mel2lin(double mel) {
    return 700.0 * (Math.exp(mel / 1127.0) - 1.0);
  }

  // Default constructor - Use 40 mel spaced bins - get sampling rate
  // and frameSize from MEAPUtil
  public AvgMelSpec() {
      this(MEAPUtil.frameSize/2+1, 40, MEAPUtil.samplingRate);
  }

  public AvgMelSpec(int N, int outDim, double sampleRate) {
    this.N = N;
    this.outDim = outDim;
    linSpec = new double[N];

    // Calculate the locations of the bin centers on the mel scale and
    // as indices into the input vector
    centersMel = new double[outDim+2];
    centersInd = new int[outDim+2];
    double melMin = lin2mel(0);
    double melMax = lin2mel(1 / (2*sampleRate));
    for(int i=0; i<outDim+2; i++) {
      centersMel[i] = melMin + i * (melMax - melMin) / (outDim + 1);
      
      // Indices in inBuf at the smallest integer greater than the mel
      // triangle peaks
      centersInd[i] = (int) (mel2lin(centersMel[i])*N*sampleRate*2 + 1);
    }

    // precalculate mel translations of fft bin frequencies
    freq2mel = new double[N];
    for(int i=0; i<N; i++)
      freq2mel[i] = lin2mel(i / (2*N * sampleRate));
  }


  public double[] features(STFT stft, long startFrame, int length) {
    double[] melSpec = new double[outDim];
    double[] curFrame;
    double sum = 0;

    // intialize average to 0
    Arrays.fill(linSpec, 0);

    // collect average linear spectrum
    for(int frame=0; frame<length; frame++) {
      curFrame = stft.getFrame(startFrame+frame);
      for(int band=0; band<linSpec.length; band++)
        linSpec[band] += curFrame[band] / length;
    }

    // convert log magnitude to linear magnitude for binning
    for(int band=0; band<linSpec.length; band++)
      //linSpec[band] = Math.exp(linSpec[band]);
      linSpec[band] = Math.pow(10,linSpec[band]/10);

    // convert to mel scale
    for(int bin=0; bin<outDim; bin++) {
      // initialize
      melSpec[bin] = 0;
      
      // ramp up
      for(int i=centersInd[bin]; i<centersInd[bin+1]; i++) 
	melSpec[bin] += linSpec[i] * (freq2mel[i] - centersMel[bin]) 
	  / (centersMel[bin+1] - centersMel[bin]);
      
      // ramp down
      for(int i = centersInd[bin+1]; i < centersInd[bin+2]; i++)
	melSpec[bin] += linSpec[i] * (centersMel[bin+2] - freq2mel[i]) 
	  / (centersMel[bin+2] - centersMel[bin+1]);

      // Take log
      melSpec[bin] = 10*Math.log(melSpec[bin]) / Math.log(10);

      sum += melSpec[bin];
    }

    // Audio scrubber takes care of normalization, level is a good cue
//     for(int bin=0; bin<outDim; bin++)
//       melSpec[bin] = melSpec[bin] / sum;
    
    return melSpec;
  }

  public String description()
  {
	  return "Computes the mean spectrum of a chunk and converts it to the perceptually weighted Mel frequency scale";
  }

}
